:- consult(cosim_semantics).

% Begin graph operations.

add_vertex(GraphIn, V, GraphIn):-
	member(V-_, GraphIn).
add_vertex(GraphIn, V, [V-[] | GraphIn]):-
	\+member(V-_, GraphIn).

del_vertex([], _, []).
del_vertex([V-_ | GraphIn], V, GraphOut):-
	del_vertex(GraphIn, V, GraphOut).
del_vertex([N-T | GraphIn], V, [N-Ts | GraphOut]):-
	delete(T, V, Ts),
	del_vertex(GraphIn, V, GraphOut).

del_vertices(GraphIn, [], GraphIn).
del_vertices(GraphIn, [V | Vertices], GraphOut):-
	del_vertex(GraphIn, V, GraphWithoutV),
	del_vertices(GraphWithoutV, Vertices, GraphOut).

merge_edge([], S-T, [S-[T]]).
merge_edge([S-Ts | GraphIn], S-T, [S-TMs | GraphIn]):-
	remove_duplicates([T|Ts], TMs).
merge_edge([N-Ts | GraphIn], S-T, [N-Ts | GraphOut]):-
	merge_edge(GraphIn, S-T, GraphOut).

add_edge(Graph, S-T, Graph):-
	member(S-Ts, Graph),
	member(T, Ts).
add_edge(GraphIn, S-T, GraphOut):-
	add_vertex(GraphIn, S, GraphS),
	add_vertex(GraphS, T, GraphT),
	merge_edge(GraphT, S-T, GraphOut).


ancestors(_, [], []).
ancestors(E, [A-T | Graph], [A | Ancestors]):-
	member(E, T),
	ancestors(E, Graph, Ancestors).
ancestors(E, [_-T | Graph], Ancestors):-
	\+member(E, T),
	ancestors(E, Graph, Ancestors).

find_all_sources(_, [], []).
find_all_sources(Graph, [S-_ | GraphAux], [S | Sources]):-
	ancestors(S, Graph, []),
	find_all_sources(Graph, GraphAux, Sources).
find_all_sources(Graph, [_ | GraphAux], Sources):-
	find_all_sources(Graph, GraphAux, Sources).

find_all_sources(Graph, Sources):-
	find_all_sources(Graph, Graph, Sources).

all_top_sort([], []).
all_top_sort(Graph, [Sources | OpsOut]):-
	find_all_sources(Graph, Sources),
	del_vertices(Graph, Sources, TrimmedGraph),
	all_top_sort(TrimmedGraph, OpsOut).

% End graph operations.

add_vertices_inports(Graph, _, [], Graph).
add_vertices_inports(GraphIn, FmuName, [port(Name,_) | PP], GraphOut):-
	add_vertices_inports(GraphIn, FmuName, PP, GraphRest),
	add_vertex(GraphRest, setIn(FmuName,Name), GraphOut).
add_vertices_outports(Graph, _, [], Graph).
add_vertices_outports(GraphIn, FmuName, [port(Name,_) | PP], GraphOut):-
	add_vertices_outports(GraphIn, FmuName, PP, GraphRest),
	add_vertex(GraphRest, getOut(FmuName,Name), GraphOut).

add_edges_inport(GraphRest, FmuName, port(Name, reactive), GraphOut):-
	add_edge(GraphRest, setIn(FmuName,Name)-doStep(FmuName), GraphOut).
add_edges_inport(GraphIn, FmuName, port(Name, delayed), GraphOut):-
	add_edge(GraphIn, doStep(FmuName)-setIn(FmuName,Name), GraphOut).
add_edges_doStep_inports(GraphIn, _, [], GraphIn).
add_edges_doStep_inports(GraphIn, FmuName, [P | PP], GraphOut):-
	add_edges_doStep_inports(GraphIn, FmuName, PP, GraphRest),
	add_edges_inport(GraphRest, FmuName, P, GraphOut).

add_edges_doStep_outports(GraphIn, _, [], GraphIn).
add_edges_doStep_outports(GraphIn, FmuName, [port(Name,_) | PP], GraphOut):-
	add_edges_doStep_outports(GraphIn, FmuName, PP, GraphRest),
	add_edge(GraphRest, doStep(FmuName)-getOut(FmuName,Name), GraphOut).

add_edges_doStep(GraphIn, Name, Inports, OutPorts, GraphOut):-
	add_edges_doStep_inports(GraphIn, Name, Inports, GraphInEdges),
	add_edges_doStep_outports(GraphInEdges, Name, OutPorts, GraphOut).

add_edges_IO_deps(GraphIn, _, _, [], GraphIn).
add_edges_IO_deps(GraphIn, FmuName, OutName, [InName | Deps], GraphOut):-
	add_edges_IO_deps(GraphIn, FmuName, OutName, Deps, GraphRest),
	add_edge(GraphRest, setIn(FmuName,InName)-getOut(FmuName,OutName), GraphOut).

add_edges_IO(GraphIn, _, [], GraphIn).
add_edges_IO(GraphIn, FmuName, [port(Name,Deps) | OutPorts], GraphOut):-
	add_edges_IO(GraphIn, FmuName, OutPorts, GraphRest),
	add_edges_IO_deps(GraphRest, FmuName, Name, Deps, GraphOut).

generate_constraint_graph_fmu([], []).
generate_constraint_graph_fmu([fmu(Name,Inports,OutPorts)|FF], Graph):-
	generate_constraint_graph_fmu(FF, GraphRest),
	add_vertex(GraphRest, doStep(Name), GraphFmuNode),
	add_vertices_inports(GraphFmuNode, Name, Inports, GraphFmuInputs),
	add_vertices_outports(GraphFmuInputs, Name, OutPorts, GraphFmuOutputs),
	add_edges_doStep(GraphFmuOutputs, Name, Inports, OutPorts, GraphFmuStateEdges),
	add_edges_IO(GraphFmuStateEdges, Name, OutPorts, Graph).

add_constraint_connections(GraphIn, _, [], GraphIn).
add_constraint_connections(GraphIn, FMUs, [connect(FO, O, FI, I) | Connections], GraphOut):-
	add_constraint_connections(GraphIn, FMUs, Connections, GraphRest),
	add_edge(GraphRest, getOut(FO,O)-setIn(FI,I), GraphOut).

generate_constraint_graph_step(FMUs, Connections, Graph):-
	generate_constraint_graph_fmu(FMUs, GraphFmus),
	add_constraint_connections(GraphFmus, FMUs, Connections, Graph).

remove_step_vertices([], Graph, Graph).
remove_step_vertices([fmu(Name,_,_)|FF], GraphIn, GraphOut):-
	del_vertex(GraphIn, doStep(Name), GraphAux),
	remove_step_vertices(FF, GraphAux, GraphOut).

generate_constraint_graph_init(FMUs, Connections, Graph):-
	generate_constraint_graph_step(FMUs, Connections, GraphStep),
	remove_step_vertices(FMUs, GraphStep, Graph).

convert_bundles_to_schedule([], []).
convert_bundles_to_schedule([B | BundleOps], Ops):-
	convert_bundles_to_schedule(BundleOps, OpsRest),
	append(B, OpsRest, Ops).

generate_step_opt_bundled(FMUs, Connections, OpsOut):-
	generate_constraint_graph_step(FMUs, Connections, Graph),
	all_top_sort(Graph, OpsOut),
	convert_bundles_to_schedule(OpsOut, Ops),
	isStepSchedule(Ops, FMUs, Connections).

generate_init_opt_bundled(FMUs, Connections, OpsOut):-
	generate_constraint_graph_init(FMUs, Connections, Graph),
	all_top_sort(Graph, OpsOut),
	convert_bundles_to_schedule(OpsOut, Ops),
	isInitSchedule(Ops, FMUs, Connections).
